namespace Syntriax.Engine.Core;

/// <summary>
/// Represents a line equation in the form y = mx + b.
/// </summary>
/// <param name="slope">The slope of the line.</param>
/// <param name="offsetY">The y-intercept of the line.</param>
/// <remarks>
/// Initializes a new instance of the <see cref="Line2DEquation"/> struct with the specified slope and y-intercept.
/// </remarks>
[System.Diagnostics.DebuggerDisplay("y = {Slope}x + {OffsetY}")]
public readonly struct Line2DEquation(float slope, float offsetY)
{
    /// <summary>
    /// The slope of the line equation.
    /// </summary>
    public readonly float Slope = slope;

    /// <summary>
    /// The y-intercept of the line equation.
    /// </summary>
    public readonly float OffsetY = offsetY;

    /// <summary>
    /// Resolves the y-coordinate for a given x-coordinate using the line equation.
    /// </summary>
    /// <param name="lineEquation">The line equation to resolve.</param>
    /// <param name="x">The x-coordinate for which to resolve the y-coordinate.</param>
    /// <returns>The y-coordinate resolved using the line equation.</returns>
    public static float Resolve(Line2DEquation lineEquation, float x) => lineEquation.Slope * x + lineEquation.OffsetY; // y = mx + b

    /// <summary>
    /// Checks if two line equations are approximately equal.
    /// </summary>
    /// <param name="left">The first line equation to compare.</param>
    /// <param name="right">The second line equation to compare.</param>
    /// <param name="epsilon">The epsilon range.</param>
    /// <returns>True if the line equations are approximately equal; otherwise, false.</returns>
    public static bool ApproximatelyEquals(Line2DEquation left, Line2DEquation right, float epsilon = float.Epsilon)
        => left.Slope.ApproximatelyEquals(right.Slope, epsilon) && left.OffsetY.ApproximatelyEquals(right.OffsetY, epsilon);

    /// <summary>
    /// Determines whether the specified object is equal to the current <see cref="Line2DEquation"/>.
    /// </summary>
    /// <param name="obj">The object to compare with the current <see cref="Line2DEquation"/>.</param>
    /// <returns><see cref="true"/> if the specified object is equal to the current <see cref="Line2DEquation"/>; otherwise, <see cref="false"/>.</returns>
    public override bool Equals(object? obj) => obj is Line2DEquation lineEquation && Slope.Equals(lineEquation.Slope) && OffsetY.Equals(lineEquation.OffsetY);

    /// <summary>
    /// Generates a hash code for the <see cref="Line2DEquation"/>.
    /// </summary>
    /// <returns>A hash code for the <see cref="Line2DEquation"/>.</returns>
    public override int GetHashCode() => System.HashCode.Combine(Slope, OffsetY);

    /// <summary>
    /// Converts the <see cref="Line2DEquation"/> to its string representation.
    /// </summary>
    /// <returns>A string representation of the <see cref="Line2DEquation"/>.</returns>
    public override string ToString() => $"{nameof(Line2DEquation)}({Slope}, {OffsetY})";
}

/// <summary>
/// Provides extension methods for the LineEquation struct.
/// </summary>
public static class Line2DEquationExtensions
{
    /// <inheritdoc cref="Line2DEquation.Resolve(Line2DEquation, float)" />
    public static float Resolve(this Line2DEquation lineEquation, float x) => Line2DEquation.Resolve(lineEquation, x);

    /// <inheritdoc cref="Line2DEquation.ApproximatelyEquals(Line2DEquation, Line2DEquation, float)" />
    public static bool ApproximatelyEquals(this Line2DEquation left, Line2DEquation right, float epsilon = float.Epsilon) => Line2DEquation.ApproximatelyEquals(left, right, epsilon);
}
